#ifndef _SEARCHENGINE_H_
#define _SEARCHENGINE_H_

#include <Common/SEGlobal.h>
#include <Tokenizers/Tokenizer.h>
#include <Matchers/Matcher.h>
#include <Rerankers/Reranker.h>

/** 
 * Our vector type in the vector-space
 * This is actually a map that holds only non-zero terms and
 * the length of the vector
 */
class tSpaceVector : public Map<std::string, double> {
public:
	// The length of the this vector
	double vLength;
	// Scalar *
	double operator*(const tSpaceVector& v) const;
	// Calculate the cosine between to vectors
	double cosine(const tSpaceVector& v) const;
};

typedef std::string tDocURI;
typedef Map<tDocID, UINT> tRow;

class SearchEngine {
	static SearchEngine* m_instance;

	typedef Map<tTerm, tRow*> tTermByDocMap;
	typedef UINT tElement;
	typedef Vector<std::string> tQueryTerms;
public:
	static SearchEngine& instance() {
		return *m_instance;
	}

	static SearchEngine* setInstance(SearchEngine* instance) {
		SearchEngine* oldInstance = m_instance;
		m_instance = instance;
		return oldInstance;
	}

	// Mapping of document internal ID to its URI
	Vector<tDocURI> m_docID2URI;
	
	// The inverse map map: tTerm->tRow
	tTermByDocMap m_termByDoc;

	void printTermByDocMap(std::ostream& s) const;

	/**
	 * Returns the row from the termByDocuments table
	 * for the supplied term.
	 * @param term - The term for which row to return
	 */
	const tRow* getTermDocuments(const tTerm& term) const;

	const UINT getNumTermDocuments(const tTerm& term) const {
		const tRow* row = getTermDocuments(term);
		if (row == NULL)
			return 0;
		return row->size();
	}

	/**
	 * Gather the document vector from our terms by documents map
	 * Scoring is done by LF.IDF
	 * @param docID   - The document vector to return by its ID
	 * @param (out) d - The space vector for the document to gather
	 */
	void getDocumentVector(UINT docID, tSpaceVector& d) const;

	Tokenizer* m_tokenizer;
	Matcher* m_matcher;
	Reranker* m_reranker;
	
public:
	SearchEngine(Tokenizer* tokenizer, Matcher* matcher, Reranker* reranker) 
		: m_tokenizer(tokenizer), m_matcher(matcher), m_reranker(reranker) {}

	/**
	 * Returns the number of indexed documents
	 */
	UINT numIndexedDocs() const { return m_docID2URI.size(); }

	/**
	 * Indexes the specified document stream into the collection
	 * @param s      - Document stream
	 * @param docURI - Unique identifier (name) for the document
	 * @return The number of terms indexed from the document
	 */
	UINT index(std::istream& s, const tDocURI& docURI);

	void removeIndexedDocument(tDocID docID);

	/**
	 * Queries the collection using the specified query string.
	 * @param queryStr - The query string
	 * @param result - The result vector where to store results URIs
	 * @return The total number of matching documents found (result may be
	 *         chopped at some point by the Matcher)
	 */
	UINT query(const std::string& queryStr, Vector<tRerankResult>& result) const;
	
	/**
	 * Queries the collection using the specified query string, while returing
     * the parsed query terms (i.e., after running the query through the tokenizer)
	 * @param queryStr  - The query string
	 * @param result - The result vector where to store results URIs
	 * @param parsedQueryTerms - The target vector for the parsed query terms
	 * @return The total number of matching documents found (result may be
	 *         chopped at some point by the Matcher)* 
	 */
	UINT query(const std::string& queryStr, Vector<tRerankResult>& result,
			   Vector<tTerm>& parsedQueryTerms) const;
		
	/**
	 * Return the document URI that is associated with the specified docID
	 */
	tDocURI getDocURIByID(tDocID docID) { return m_docID2URI[docID]; }

	/**
	 * Stores the indexed collection to disk
	 */
	bool storeCollection(const std::string& filename) const;

	/**
	 * Loads the indexed collection from disk.
	 * Any trensient indexed data is lost.
	 * @return The number of indexed documents retrieved
	 */
	bool loadCollection(const std::string& filename);
};

#endif // _SEARCHENGINE_H_