#include <Common/getopt.h>
#include <Engine/SearchEngine.h>

#include <Tokenizers/WordTokenizer.h>
#include <Tokenizers/LovinsWordTokenizer.h>
#include <Tokenizers/PorterWordTokenizer.h>
#include <Tokenizers/PaiceWordTokenizer.h>
#include <Tokenizers/StopWordTokenizer.h>

#include <Matchers/AnyMatcher.h>
#include <Matchers/AllMatcher.h>
#include <Rerankers/TFIDFReranker.h>

#include <algorithm>

#define COL_FNAME_PREFIX "collection"
#define COL_FNAME_SUFFIX ".col"

// The HTML_OUT macro controls makes output
// in html readable text
#ifdef HTML_OUT
#define END_L "<br>\n"
#define GET_OPT_STRING "q:s:m:nh:"
#else /* console */
#define END_L "\n"
#define GET_OPT_STRING "q:s:m:n"
#endif /* HTML_OUT */

void usage(const std::string& msg = "") {
	if (msg.size() > 0) {
		std::cout << "Error: " << msg << END_L;
	}
#ifdef HTML_OUT
	std::cout << "<u>";
#endif
	std::cout << "SE: Search Engine implementation"
#ifdef HTML_OUT
	    << "</u>"
#endif
	    << END_L
#ifndef HTML_OUT
		<< "--------------------------------" << END_L
#endif /* HTML_OUT */
		<< "Usage: SE [options] file1 file2 ... " << END_L
		<< " file1 ... - Are the files to index (may be empty)" << END_L
		<< " options:" << END_L
		<< "   -q queryString              : Run the supplied query on the collection" << END_L
		<< "   -s porter/paice/lovins/none : Use the given stemmer (default is porter)" << END_L
		<< "   -m all/any                  : Match any query token (default) or all query tokens" << END_L
		<< "   -n                          : Do not ignore stopwords (default is to ignore stopwords)" << END_L
#ifdef HTML_OUT
		<< "   -h [linkPrefix]             : Print results in HTML format prefixing links with optional arg" << END_L
#endif
		<< "Examples:" << END_L 
		<< "  SE lecture1.txt lecture2.txt" << END_L
		<< "\t- This will index the specified files in the collection" << END_L 
		<< "  SE -q 'IR LF IDF'" << END_L
		<< "\t- This will query the collection for IR and/or LF and/or IDF" << END_L
		<< "  SE -s none -n -m any -q 'vector space model'" << END_L
		<< "\t- This will query the no-stemmer no-stopword collection for documents" << END_L
		<< "\t  with all the query terms (i.e, vector & space & model)" << END_L
		<< "Note: Each configuration uses a different indexing collection file (.col)" << END_L;
}

int main(int argc, char** argv) {

	try {

	if (argc == 1) {
		usage();
		return -1;
	}

	std::string queryStr = "";
	std::string stemmerStr = "porter";
	std::string matcherStr = "any";
	std::string linkPrefix = "";
	bool ignoreStopWords = true;
	
	int c;
	while ( (c = getopt(argc, argv, GET_OPT_STRING)) != -1) {
		switch (c) {
			case 'q': 
				queryStr = optarg;
				break;
			case 's':
				stemmerStr = optarg;
				std::transform(stemmerStr.begin(), stemmerStr.end(), stemmerStr.begin(), tolower);
				break;
			case 'm':
				matcherStr = optarg;
				std::transform(matcherStr.begin(), matcherStr.end(), matcherStr.begin(), tolower);
				break;
#ifdef HTML_OUT
			case 'h':
				linkPrefix = optarg;
				break;
#endif HTML_OUT
			case 'n':
				ignoreStopWords = false;
				break;
			case ':':
			case '?':
				usage();
				return -1;
		}
	}

	// construct a chained tokenizer according to the stopword param
	Tokenizer *tokenizer;
	Tokenizer *chainedTokenizer;

	if (ignoreStopWords)
		chainedTokenizer = new StopWordTokenizer();
	else
		chainedTokenizer = new WordTokenizer();

	// Construct tokenizer according to the stemmer param

	if (stemmerStr == "none") {
		tokenizer = chainedTokenizer;
	} else if (stemmerStr == "porter") {
		tokenizer = new PorterWordTokenizer(chainedTokenizer);
	} else if (stemmerStr == "paice") {
		tokenizer = new PaiceWordTokenizer(chainedTokenizer);
	} else if (stemmerStr == "lovins") {
		tokenizer = new LovinsWordTokenizer(chainedTokenizer);
	} else {
		usage("Unsupported stemmer '" + stemmerStr + "'");
	}

	// Construct the matcher according to the matcher param
	Matcher* matcher;
	if (matcherStr == "any") {
		matcher = new AnyMatcher(100);
	} else if (matcherStr == "all") {
		matcher = new AllMatcher(100);
	} else {
		usage("Unsupported matcher '" + matcherStr + "'");
	}
	
	// Construct the reranker
	Reranker* reranker = new TFIDFReranker();

	// Set the singelton SE instance
	SearchEngine::setInstance(new SearchEngine(tokenizer, matcher, reranker));

	// Generate the collection filename from the parameters (different collection per configuration)
	std::string colFilename = std::string(COL_FNAME_PREFIX) + "_" + stemmerStr + (ignoreStopWords ? "_stopwords" : "_nostopwords") + COL_FNAME_SUFFIX;
	
	std::cout << "Loading old collection " << colFilename << " ... " << std::flush;
	if (!SearchEngine::instance().loadCollection(colFilename)) {
		std::cout << "Failed (either no old collection or corrupt collection)" << END_L;
	} else {
		std::cout << "Done [" << SearchEngine::instance().numIndexedDocs() << " document(s) indexed]" << END_L;
	}

	bool indexedNewDocs = false;
	// Index the supplied files
	for (int i = optind; i < argc; i++) {
		const char* filename = argv[i];
		std::cout << "Indexing " << filename << " ... " << std::flush;
		std::ifstream f(filename);
		if (f.is_open()) {
			UINT numTokens = SearchEngine::instance().index(f, filename);
			std::cout << "Done [" << numTokens << " token(s) indexed]" << END_L;
			indexedNewDocs = true;
		} else {
			std::cout << "Failed, file not found" << END_L;
		}
	}
	// Store the collection if it has changed
	if (indexedNewDocs) {
		std::cout << "Storing updated collection ... " << std::flush;
		SearchEngine::instance().storeCollection(colFilename);
		std::cout << "Done" << END_L;
	}

	// Query the collection
	if (queryStr.size() > 0) {
		if (SearchEngine::instance().numIndexedDocs() == 0) {
			std::cout << "Can't query an empty collection" << END_L;
			return -1;
		}
		Vector<tRerankResult> result;
		std::cout << "Quering '" << queryStr << "' ... " << std::flush;
		Vector<tTerm> parsedQueryTerms;
		UINT numMatched = SearchEngine::instance().query(queryStr, result, parsedQueryTerms);
		std::cout << "Done [" << numMatched << " document(s) matched]" << END_L 
			<< "Tokenized query: " << parsedQueryTerms << END_L << END_L;
		std::cout << "#terms" << END_L;
#ifdef HTML_OUT
		std::cout << "<table><tr><u><td>matched</td><td>Matched document URI</td><td>rerankScore</td></u></tr>" << std::endl;
#else 
		printf("%-8s %-35s (%s)\n", "matched", "Document URI (filename)", "rerank score");
		std::cout << "-----------------------------------------------------------\n";
#endif /* HTML_OUT */
		for (int i = 0; i < result.size(); i++) {
			tDocURI docURI = SearchEngine::instance().getDocURIByID(result[i].docID);
#ifdef HTML_OUT
			std::cout << "<tr><td>" << (int)result[i].matchScore << "</td>" 
			          << "<td><a href=\"" << linkPrefix << docURI << "\">" << docURI << "</a></td>"
					  << "<td>" << result[i].rerankScore << "</td></tr>" << std::endl;
#else
			printf("%-8.0lf %-35s (%lf)\n", result[i].matchScore, docURI.c_str(), result[i].rerankScore);
#endif /* HTML_OUT */
		}
#ifdef HTML_OUT
		std::cout << "</table>\n";
#endif
	} else {
		// Empty query string
		std::cout << "Not quering (no query string supplied)" << END_L;
	}

	} catch (SEException e) {
		std::cout << "Caugh " << e << END_L;
		return -1;
	} catch (...) {
		std::cout << "Caught non-SEException, bailing out" << END_L;
		return -1;
	}

	return 0;
}
